inlets = 3;
outlets = 2;

// Global Constants
var CONTROL_SURFACE = "control_surfaces 0 ";

// Listener Arrays
var buttonShift;
var buttonArm = [];
var buttonCrossfaderSelect = [];
var buttonSolo = [];
var buttonMute = [];
var buttonCL = [];
var buttonMFX = [];
var buttonLoopStop = [];
var buttonSL = [];
var buttonTrackActivate = [];
var buttonFXReset = [];
var listenerTracks = [];
var listenerClipLooping = [];
var listenerClipPlaying = [];
var defaultsEffects = [];

// Global Dynamics
var boolEnabled = 0;				// Checked by the init() function to determine if the template setup has taken place or not. Prevents duplicate instances.
var boolCue = 0;					// Used to see if using the Track Select button will cue/solo a track or not (good to have for soundcards without cue outs)

// Random usage variable
var lAPI;
var i;
var j;
var t;
var intArray;

function msg_int() {
	if (inlet == 2) {
		boolCue = arrayfromargs(arguments)[0];
		post ("Cueing set to: " + boolCue + "\n");
	}
}

function bang() {

	switch (inlet) {
		case 0:	// Bang to start the script
			if (!boolEnabled) {
				init();
			}
			break;
		case 1:
			if (boolEnabled) {
				for (i = 0; i <= 7; i++) {	// Run through the MFX, see what's pressed, and bump the parameter as needed.
					if (buttonMFX[i].ON == 1) {
						//post("Bang produced a result: bumping " + i + "\n");
						MFXBump(i);
					}
				}
			}
			break;
		default:
			break;
	}
}

// function anything()
// Catches any weirdness from the patcher being sent to this script.

function anything() {
	post("Anything Received: " + arrayfromargs(arguments) + " on Inlet #" + inlet + "\n");
}

// function path()
// Listens for the path argument from the m4l.api.selectcontrolsurface patcher
// that allows the user to pull their specific control surface. This kicks off 
// the show by making sure the name matches "APC40_MkII" and running the init()
// function

function path() {
	switch (inlet) {
		case 0:
			CONTROL_SURFACE = arrayfromargs(arguments)[0] + " " + arrayfromargs(arguments)[1] + " ";
			post("Control Surface Selected.\n");
			break;
		default:
			break;
	}
}


function init() {

	lAPI = new LiveAPI();
	
	lAPI.path = CONTROL_SURFACE;
	
	if (lAPI.id != 0) {
		lAPI.id = 0;					// Free up lAPI
	
		LightShow();					// Light show for startup sequence.
	
		disableSLButtons();			
		setupSLButtons();
		disableRightGrid();
		disableTrackSelectButtons();
		SetupLoopControlButtons();
		setupCLGrid();
		setupMFXGrid();
		setupLoopToggleButtons();
		setupClipLoopingListeners();
		setupTrackListeners();
		setupEffectResetButtons();
		disableDeviceControlButtons();
		setupDropRepeatButtons();
		SetupShiftButton();
		outlet(0, "bang");				// Starts the metro object
		outlet(1, 1);					// Sets the default for cue/solo to being on.

		boolEnabled = 1;				// Ensures no duplicate instances are spawned.

		post("Initialized.\n");
	} else {
		lAPI.id = 0;
		post("Invalid control surface selected. Exiting.\n");
	}
}

function LightShow() {
	for(j = 22; j <= 61; j ++) {
		lAPI.path = CONTROL_SURFACE + "controls " + j;
		lAPI.call("send_value", 67);
		lAPI.call("send_value", 21, 0, 9);
	}

	lAPI.id = 0;	
}

// function setupTrackListeners()
// This function sets up the variable array for listenerTracks[] that will track playing status of a track (0 - 7) to 
// handle loop lighting and perhaps other feedback down the line

function setupTrackListeners() {
	for(i = 0; i <= 7; i++) {
		listenerTracks[i] = new LiveAPI(TrackListenerCallback, "live_set tracks " + i);
		listenerTracks[i].property = "playing_slot_index";
		listenerTracks[i].TrackID = i;
	}
	post("Track Listeners Setup Successfully.\n");
}

// function TrackListenerCallback(args)
// Callback function to handle a change in playing_slot_index on a track
// This will handle looping feedback and perhaps more down the line

function TrackListenerCallback(args) {
	if (args[0] == "playing_slot_index") {
		switch(this.TrackID) {
			case 0:
			case 1:
			case 2:
			case 3:
				if (args[1] >= 0) {
					listenerClipLooping[this.TrackID].path = "live_set tracks " + this.TrackID + " clip_slots " + args[1] + " clip";	// Assign appropriate clip loop listener to the playin clip
					listenerClipLooping[this.TrackID].property = "looping";	// Assign looping as the property to listen for
				} else {
					listenerClipLooping[this.TrackID].id = 0;	// Wipe out the clip loop listener
					listenerClipLooping[this.TrackID].IsLooping = 0;	// Set it's looping boolean to 0
					buttonLoopStop[this.TrackID].call("send_value", buttonLoopStop[this.TrackID].LightOFF);	// Turn off the loop indicator/toggle button
				}
				break;
			case 4:
			case 5:
			case 6:
			case 7:
				if (args[1] >= 0) {
					i = (this.TrackID - 4) + (args[1] * 4);	// Simplify calculation usage
					buttonCL[i].call("send_value", buttonCL[i].LightPLAYING, 0, 10);
					buttonCL[i].PlayState = 2;	// Playing
					if (i <= 3) {
						buttonCL[i + 4].call("send_value", buttonCL[i + 4].LightSTOPPED);
						buttonCL[i + 4].PlayState = 0;	// Stopped
					} else {
						buttonCL[i - 4].call("send_value", buttonCL[i - 4].LightSTOPPED);
						buttonCL[i - 4].PlayState = 0;	// Stopped
					}
				} else {
					buttonCL[this.TrackID - 4].call("send_value", buttonCL[this.TrackID - 4].LightSTOPPED);
					buttonCL[this.TrackID].call("send_value", buttonCL[this.TrackID].LightSTOPPED);
				}
				break;
			default:
				break;
		}
	}	
}

// function setupClipLoopingListeners()
// This function will setup the listeners for looping clips. It is reliant on the setupTrackListeners() function
// for path setting as the set is dynamic and doesn't occupy a fixed space.

function setupClipLoopingListeners() {
	for(i = 0; i <= 3; i++) {
		listenerClipLooping[i] = new LiveAPI(ClipLoopingCallback)
		listenerClipLooping[i].TrackID = i;
		listenerClipLooping[i].IsLooping = 0;
	}
	post("Clip Looping Listeners Setup Successfully.\n");
}

// function ClipLoopingCallback(args)
//

function ClipLoopingCallback(args) {
	if (args[0] == "looping") {
		this.IsLooping = args[1];
		switch(args[1]) {
			case 0:
				buttonLoopStop[this.TrackID].call("send_value", buttonLoopStop[this.TrackID].LightOFF);
				break;
			case 1:
				buttonLoopStop[this.TrackID].call("send_value", buttonLoopStop[this.TrackID].LightON, 0, 9);				
				break;
			default:
				break;
		}
	}
}




// Shift Button
// The following code establishes the shift button listener. Default functionality is intact, however this just resets the lights
// upon coming back from a shift press.

function SetupShiftButton() {
	buttonShift = new LiveAPI(ShiftCallback, "control_surfaces 0 controls 0");
	buttonShift.property = "value";
	
	post("Shift Button Listener Setup.\n");
}

function ShiftCallback(args) {
	if (args[0] == "value") {
	
		switch(args[1]) {	
			case 0:
				post("Shift Button Released.\n");
				// Here we restore all of our pretty lights...
				
				//buttonCL[];
				//buttonMFX[];
				for(i = 0; i <= 7; i ++) {
					// Restore MFX Lighting
					buttonMFX[i].call("send_value", buttonMFX[i].LightOFF);					
					
					// Restore CL Lighting
					buttonCL[i].call("send_value", buttonCL[i].LightSTOPPED);
					switch(buttonCL[i].PlayState) {
						case 0:
							break;
						case 1:
							// buttonCL[i].call("send_value", buttonCL[i].LightTRIGGERED, 0, 14);	// We haven't checked for fire commands on buttons without a clip. Just omitting for right now
							break;
						case 2:
							buttonCL[i].call("send_value", buttonCL[i].LightPLAYING, 0, 10);
							break;
						default:
							break;
					}
				}


				//buttonLoopStop[];
				for(i = 0; i <= 3; i ++) {
					buttonLoopStop[i].call("send_value", buttonLoopStop[i].LightOFF);
					if (listenerClipLooping[i].IsLooping) {
						buttonLoopStop[i].call("send_value", buttonLoopStop[i].LightON, 0, 9);
					}
				}
				
				//SLButtons[]
				disableSLButtons();			
				break;
			case 127:
				post("Shift Button Pressed.\n");
				break;
			default:
				break;
		}	
	}	
}

// TRACK SELECT BUTTONS
// The following disable/setup function and callback handle the track select buttons. It allows the user to detail/cue the playing clip
// on any given track.
// 

function disableTrackSelectButtons() {
	// Setup track select buttons to bring up the playing clip to detail view.
	for(i = 0; i <= 7; i ++) {
		buttonTrackActivate[i] = new LiveAPI(ClipViewCallback, CONTROL_SURFACE + "controls " + (95 + i));
		buttonTrackActivate[i].TrackID = i;
		buttonTrackActivate[i].property = "value";
	}
	
}

function ClipViewCallback(args) {
	switch(args[1]) {
		case 127:
			SetDetailCue(this.TrackID);
			break;
		default:
			break;
	}
	
	lAPI.id = 0;
}


// DETAIL/CUE FUNCTION
// Used by the preceding two function sets. This little dandy will navigate to a track, find the current
// playing clip, and then focus it in detail view and solo/cue the track. I use this for setting up my decks ahead of time with
// my headphones.
//

function SetDetailCue(intTrackID) {
	lAPI.path = "live_set tracks " + intTrackID;
	if (lAPI.get("playing_slot_index") >= 0) {
		lAPI.path = "live_set tracks " + intTrackID + " clip_slots " + lAPI.get("playing_slot_index") + " clip";
		i = lAPI.id;
		lAPI.path = "live_set view";
		lAPI.set("detail_clip", "id", i);
		lAPI.path = "live_app view";
		lAPI.call("show_view", "Detail/Clip");
		if (boolCue) {
			SetCue(intTrackID);
		}
		post("Detail / Cue set on Track " + (parseInt(intTrackID) + 1) + ".\n");
	}
}

function SetCue(args) {
	for (i = 0; i <= 3; i++) {
		lAPI.path = "live_set tracks " + i;
		if (i != args) {
			lAPI.set("solo", 0);
		} else {
			lAPI.set("solo", 1);
		}
	}
}


// Looping buttons. Take control of the track activator, solo/cue, crossfade, and record
// buttons to turn them into loop buttons. 
//


function SetupLoopControlButtons() {
	// For funsies, we disable the Stop_All_Clips_Button here, too.
	lAPI.path = CONTROL_SURFACE + "components 1";
	lAPI.call("set_stop_all_clips_button", "id", 0);

	// Disables the Activator, Solo/Cue, and Record Arm buttons
	intArray = [58, 121, 130, 112, 103];
	
	// Tracks 1 through 4 Setup and Assignment of Looping Buttons and Callback Information
	
	for(i = 0; i <= 3; i ++) {
		lAPI.path = CONTROL_SURFACE + "components " + (intArray[0] + i);
		lAPI.call("set_solo_button", "id", 0);
		lAPI.call("set_crossfade_toggle", "id", 0);
		lAPI.call("set_mute_button", "id", 0);
		lAPI.call("set_arm_button", "id", 0);
		
		// Track Activator 		at 4 beats
		buttonMute[i] = new LiveAPI(LoopCallback, CONTROL_SURFACE + "controls " + (intArray[1] + i));
		buttonMute[i].TrackID = i;
		buttonMute[i].Beats = 4;
		buttonMute[i].property = "value";

		// Crossfader Selector 	at 8 beats
		buttonCrossfaderSelect[i] = new LiveAPI(LoopCallback, CONTROL_SURFACE + "controls " + (intArray[2] + i));
		buttonCrossfaderSelect[i].TrackID = i;
		buttonCrossfaderSelect[i].Beats = 8;
		buttonCrossfaderSelect[i].property = "value";

		// Solo / Cue 			at 16 beats
		buttonSolo[i] = new LiveAPI(LoopCallback, CONTROL_SURFACE + "controls " + (intArray[3] + i));
		buttonSolo[i].TrackID = i;
		buttonSolo[i].Beats = 16;
		buttonSolo[i].property = "value";

		// Record / Arm 		at 32 beats
		buttonArm[i] = new LiveAPI(LoopCallback, CONTROL_SURFACE + "controls " + (intArray[4] + i));
		buttonArm[i].TrackID = i;
		buttonArm[i].Beats = 32;
		buttonArm[i].property = "value";

	}
	
	lAPI.id = 0;
	post("Track Control Buttons (Crossfade, Solo/Cue, Activate, Record) Disabled.\n");
	post("Loop Buttons Initialized.\n");
}



function LoopCallback(args) {
	var intPP;
	
	switch (args[1]) {
		case 127:
			lAPI.path = "live_set tracks " + this.TrackID;
			if (lAPI.get("playing_slot_index") >= 0) {
				lAPI.path = "live_set tracks " + this.TrackID + " clip_slots " + lAPI.get("playing_slot_index") + " clip";
				intPP = parseInt(lAPI.get("playing_position"), 10);
				lAPI.set("looping", 1);
				if (lAPI.get("loop_start") > intPP) {
					lAPI.set("loop_start", intPP);
					lAPI.set("loop_end", intPP + this.Beats); 
				} else {
					lAPI.set("loop_end", intPP + this.Beats); 
					lAPI.set("loop_start", intPP);
				}
			}
			lAPI.id = 0;
			break;
		default:
			break;
	}
}	

// setupDropRepeatButtons()
// Function that disables and assigns functionality to the Record buttons on tracks 5-8
// as Drop Repeat momentaries for tracks 1-4

function setupDropRepeatButtons() {
	for (i = 4; i <= 7; i++) {
		lAPI.path = CONTROL_SURFACE + "components " + (58 + i);
		lAPI.call("set_arm_button", "id", 0);
		buttonArm[i] = new LiveAPI(DropRepeatCallback, CONTROL_SURFACE + "controls " + (i + 103));
		buttonArm[i].TrackID = i - 4;
		buttonArm[i].Parameter = 7;		// Knob number for Arm button momentary effect
		buttonArm[i].ParameterValue = 90;	// Effect value for arm button momentary effect
		buttonArm[i].property = "value";		
		buttonArm[i].call("turn_on");	// Turn on light as "off" indicator
	}
	lAPI.id = 0;
	post("Arm on tracks 5-8 disabled and assigned to Drop Repeat.\n");
}

// DropRepeatCallback(args)
// Callback that momentarily fires Drop Repeat with a value of 90 (1/8 note)

function DropRepeatCallback(args) {
	switch (args[1]) {
		case 127:
			lAPI.path = "live_set tracks " + this.TrackID + " devices 0 parameters " + this.Parameter;
			lAPI.set("value", this.ParameterValue);
			this.call("turn_off");
			break;
		case 0:
			lAPI.path = "live_set tracks " + this.TrackID + " devices 0 parameters " + this.Parameter;
			lAPI.set("value", 0);
			this.call("turn_on");
			break;
		default:
			break;
	}
	lAPI.id = 0;
}


// The following functions disable the Device Control Buttons (bottom eight buttons on APC40)
// We use the Device <- / Device ->, Bank <- / Bank -> buttons as effect resets for each respective
// deck #'s 1-4
//

function disableDeviceControlButtons() {
	// Disable the "Dev On/Off" and "Device Lock" Buttons
	lAPI.path = CONTROL_SURFACE + "components 70";
	lAPI.call("set_enabled", 0);

	// The next two lines disable the "Clip/Track", "Detail View", "<-", and "->" buttons
	lAPI.path = CONTROL_SURFACE + "components 71";
	lAPI.call("set_enabled", 0);

	// Post about it to brag!
	post("Bottom eight buttons disabled.\n");	
}

function setupEffectResetButtons() {
	// Create first dimension of array (4 tracks)
	defaultsEffects = new Array(4);	

	for(i = 0; i <= 3; i++) {
		// Create second dimension of array within the first
		defaultsEffects[i] = new Array(8)
		defaultsEffects[i][0] = 63;		// Knob #1 value
		defaultsEffects[i][1] = 0;		// Knob #2 value
		defaultsEffects[i][2] = 0;		// Knob #3 value
		defaultsEffects[i][3] = 0;		// Knob #4 value
		defaultsEffects[i][4] = 0;		// Knob #5 value
		defaultsEffects[i][5] = 0;		// Knob #6 value
		defaultsEffects[i][6] = 0;		// Knob #7 value
		defaultsEffects[i][7] = 0;		// Knob #8 value

		buttonFXReset[i] = new LiveAPI(FXResetCallback, CONTROL_SURFACE + "controls " + (i + 188));
		buttonFXReset[i].TrackID = i;
		buttonFXReset[i].property = "value";
		buttonFXReset[i].call("turn_on");
	}
	
	lAPI.id = 0;
	post("Effect reset buttons setup.\n");
	
}

function FXResetCallback(args) {
	switch (args[1]) {
		case 127:
			ResetFX(this.TrackID);
			break;
		default:
			break;
	}
}

function ResetFX(pTrackID) {
	for (i = 0; i <= 7; i++) {
		lAPI.path = "live_set tracks " + pTrackID + " devices 0 parameters " + (i + 1);
		lAPI.set("value", defaultsEffects[pTrackID][i]);
	}
	
	lAPI.id = 0;
}

// RIGHT GRID FUNCTIONS: DISABLE, SETUP, CALLBACK(s)
// These functions all relate to the right-most 20 buttons on the APC40 Clip Launch Grid
// 

function disableRightGrid() {
	// Disable right 20 clip launch buttons
	for(i = 16; i <= 55; i++) {
		button = new LiveAPI(CONTROL_SURFACE + "components " + i);
		button.call("set_enabled", 0);
		if ((i - 1) % 9 === 0) {
			i = i + 5;
		}
	}

	post("Right Grid Disabled.\n");

}

// FIXED CLIP LAUNCH GRID SETUP AND CALLBACK

function setupCLGrid() {
	intArray = [26, 34];

	for(j = 0; j <= 1; j++) {
		for(i = 0; i <= 3; i++) {
			buttonCL[i + (j * 4)] = new LiveAPI(ClipLaunchCallback, CONTROL_SURFACE + "controls " + (intArray[j] + i));
			lAPI.path = "live_set tracks " + (i + 4) + " clip_slots " + j;
			buttonCL[i + (j * 4)].ClipLaunchID = lAPI.id;
			buttonCL[i + (j * 4)].LightSTOPPED = 43;
			buttonCL[i + (j * 4)].LightTRIGGERED = 75;
			buttonCL[i + (j * 4)].LightPLAYING = 21;
			buttonCL[i + (j * 4)].PlayState = 0;	// 0 == STOPPED, 1 == TRIGGERED, 2 == PLAYING
			buttonCL[i + (j * 4)].property = "value";
			buttonCL[i + (j * 4)].call("send_value", buttonCL[i + (j * 4)].LightSTOPPED);
		}
		post("Deck " + j + " Fixed Clip Launch Buttons Setup.\n");
	}
	
	lAPI.id = 0;	
}

function ClipLaunchCallback(args) {
	switch (args[1]) {		// Operate based on button pressed or not (127 == pressed, 0 == not pressed)
		case 127:
			lAPI.id = parseInt(this.ClipLaunchID);
			if (parseInt(lAPI.get("has_clip"))) {
				lAPI.call("fire");
				this.call("send_value", this.LightTRIGGERED, 0, 14);
				this.PlayState = 1;
			}
			break;
		case 0:
			break
		default:
			break;		
	}	

	lAPI.id = 0;
}


// MOMENTARY FX GRID SETUP AND CALLBACK

function setupMFXGrid() {
	intArray = [42, 50];

	for(j = 0; j <= 1; j++) {
		for(i = 0; i <= 3; i++) {
			buttonMFX[i + (j * 4)] = new LiveAPI(MFXCallback, CONTROL_SURFACE + "controls " + (intArray[j] + i));
			buttonMFX[i + (j * 4)].Deck = i;
			buttonMFX[i + (j * 4)].ID = (i + (j * 4));
			buttonMFX[i + (j * 4)].LightON = 9;
			buttonMFX[i + (j * 4)].LightOFF = 89;
			if (j == 0) {
				buttonMFX[i + (j * 4)].Parameter = 5;	// Parameter controlled by top MFX button
			} else {
				buttonMFX[i + (j * 4)].Parameter = 4;	// Parameter controlled by bottom MFX button
			}
			buttonMFX[i + (j * 4)].ON = 0;
			buttonMFX[i + (j * 4)].property = "value";
			buttonMFX[i + (j * 4)].call("send_value", buttonMFX[i + (j * 4)].LightOFF);
		}
		post("MFX Deck " + j + " Setup Successfully\n");
	}
	lAPI.id = 0;
}


function MFXCallback(args) {
	if (args[0] == "value") {
		switch (args[1]) {
			case 127:
				this.call("send_value", this.LightON, 0, 10);
				this.ON = 1;
				break;
			case 0:
				this.call("send_value", this.LightOFF);
				this.ON = 0;
				lAPI.path = "live_set tracks " + this.Deck + " devices 0 parameters " + this.Parameter;
				lAPI.set("value", 0);
				break;
			default:
				break;
		}	
	}
}

function MFXBump(pID) {
	lAPI.path = "live_set tracks " + buttonMFX[pID].Deck + " devices 0 parameters " + buttonMFX[pID].Parameter;
	i = parseInt(lAPI.get("value"));

	if (i == 0) i = -1;	// Handls the whole 0-to-127 range thing to keep our math tidy.

	if (i < 127) {
		i = i + 8
		lAPI.set("value", i);
	}

	lAPI.id = 0;
}


// LOOP TOGGLES
// These function assign the bottom four of the clip launch grid to their respective channel to indicate a loop
// is in effect and allow you to toggle looping for on/off as well.
//

function setupLoopToggleButtons() {	
	// Assign button variables for loop-stoppage in the callback function above.
	for(i = 58; i <= 61; i ++) {
		buttonLoopStop[i - 58] = new LiveAPI(LoopToggleCallback, CONTROL_SURFACE + "controls " + i);
		buttonLoopStop[i - 58].TrackID = (i - 58);
		buttonLoopStop[i - 58].property = "value";
		buttonLoopStop[i - 58].LightOFF = 72;
		buttonLoopStop[i - 58].LightON = 45;
		buttonLoopStop[i - 58].call("send_value", buttonLoopStop[i - 58].LightOFF);
	}
	
	post("Loop Stop Buttons Setup.\n");
	
}

function LoopToggleCallback(args) {
	switch (args[1]) {
		case 127:
			lAPI.path = "live_set tracks " + this.TrackID;
			if (lAPI.get("playing_slot_index") >= 0) {
				lAPI.path = "live_set tracks " + this.TrackID + " clip_slots " + lAPI.get("playing_slot_index") + " clip";
				if (lAPI.get("looping") == 1) {	
					lAPI.set("looping", 0);
				} else {
					lAPI.set("looping", 1);
				}
			}
			lAPI.id = 0;
			break;
		default:
			break;
	}
}


// function disableSLButtons()
// Disables Scene Launch buttons on the MKII

function disableSLButtons() {
	for (i = 11; i <= 47; i = i + 9) {
		lAPI.path = CONTROL_SURFACE + "components " + i;
		lAPI.call("set_launch_button", "id", 0);
		lAPI.id = 0;
	}

	post("Scene Launce buttons disabled.\n");	
}

// function setupSLButtons()
// Builds out the buttonSL[] array with callback SLCallback(args)

function setupSLButtons() {
	for (i = 0; i <= 4; i++) {
		buttonSL[i] = new LiveAPI(SLCallback, CONTROL_SURFACE + "controls " + (i + 16));
		buttonSL[i].SceneID = i;
		buttonSL[i].property = "value";
		buttonSL[i].call("send_value", 0);
		buttonSL[i].call("send_value", 44, 0, 10);
	}

}

// function SLCallback(args)
// Callback function for scene launch buttons

function SLCallback(args) {
	switch(args[1]) {
		case 127:
			buttonSL[this.SceneID].call("send_value", 25, 0, 9);
			//post("Scene Launch " + this.SceneID + " Button Pressed.\n");
			break;
		case 0:
			buttonSL[this.SceneID].call("send_value", 44, 0, 10);
			//post("Scene Launch " + this.SceneID + " Button Released.\n");
			break;
		default:
			break;
	}
	
}

